---
title: Complete binary trees
nav_order: 0
---

When we introduced 2-3 trees, we first proposed the following invariant: each node must have either 0 or 2 children. (A binary tree with this property is known as a **full binary tree**.) But we also came up with a worst-case, spindly binary tree structure that maintained this invariant.

![Worst-Case Full Tree](https://docs.google.com/drawings/d/e/2PACX-1vQg6teoRmSeiw10PybR0wxaTVUVr9hW5Lb56BJO48ly2u0KSilPH_BZbH32sXV0DHnchXTCQYD925ZS/pub?w=960&h=540)

This led us to consider a different hypothesis: that unbalanced growth leads to worst-case height trees. This hypothesis setup the foundation for 2-3 trees.

But what if we turned back time and didn't give up here? Consider the binary trees depicted below. Let's develop an invariant that includes the balanced trees and excludes unbalanced trees.

![Identifying Complete Trees](https://docs.google.com/drawings/d/e/2PACX-1vQaPhCapHaVoHy0cTLU36fPZa0oQ7ksy2geiIwTnnYcsm165s2VbVJ-2UHdLag93dlvvnsdKbsnYhfA/pub?w=960&h=540)

One such invariant is the definition of **complete binary trees**.

Complete tree
: A tree where every level (except possibly the last) is completely filled, and the last level is filled left to right.

Unfortunately, it's hard to maintain the binary search tree invariant together with the completeness invariant. Consider how we might maintain the completeness invariant while adding the item **A** to the following BST.

![Impossible Complete BST Insertion](https://docs.google.com/drawings/d/e/2PACX-1vTKpUB2VbLhMHmrm47qvxw01gxMXVV7T1H74sKq--TDCXiz_lSVITIcx4mudsw3VMqdzJpxtpzGQgKQ/pub?w=960&h=540)

Note that this isn't possible with rotation alone! Some of the nodes have to move quite a distance: the leaf **D**, which was previously the right child of **C**, becomes the overall root of the entire tree as a result of inserting **A**!

If we want completeness, then we can't have the binary search tree invariant. But without the BST invariant, we can't efficiently implement a set or map abstract data type---all of our search trees relied on a search tree invariant to help us decide, at each node, whether we would expect to find the target value in the left/middle/right subtree.
