---
title: What would a 2-3 tree do?
nav_order: 7
---

The implementation for an LLRB tree relies on the invariants and 1-1 correspondence with 2-3 trees. 2-3 tree operations like overstuffing leaves and splitting nodes can be implemented with **rotations** and **recoloring** in an LLRB tree. In any situation, we can always ask: What would a 2-3 tree do?

<div style="--aspect-ratio: 720 / 405; --add-height: 29px"><iframe src="https://docs.google.com/presentation/d/e/2PACX-1vRFcFwMtY8LjSGWLYla1XXKu8CYAKAJWihRSn5egS_YBDGNITh7RMId1zuvQEGqC0yCVxa9E4IwtWd4/embed?start=false&loop=false&delayms=3000" frameborder="0" width="720" height="434" allowfullscreen="true" mozallowfullscreen="true" webkitallowfullscreen="true"></iframe></div>

The following simplified snippet from the implementation of [RedBlackBST.java](https://algs4.cs.princeton.edu/33balanced/RedBlackBST.java.html) first performs the standard recursive BST insertion (new leaves are colored red) and then performs some maintenance when returning from each recursive call.

```java
private Node add(Node h, Key key) {
    // New leaves are colored red
    if (h == null)
        return new Node(key, RED);

    // Recursive BST insertion
    int cmp = key.compareTo(h.key);
    if (cmp < 0)
        h.left  = add(h.left,  key);
    else if (cmp > 0)
        h.right = add(h.right, key);
    else
        return h; // do not add duplicates

    // Maintain LLRB tree invariants
    if (isRed(h.right) && !isRed(h.left))
        h = rotateLeft(h);
    if (isRed(h.left) && isRed(h.left.left))
        h = rotateRight(h);
    if (isRed(h.left) && isRed(h.right))
        flip(h);

    return h;
}
```

When **returning** from each recursive call to `add`, the code addresses each of these conditions as necessary.

1. If the right child is red and the left child is black, **rotate left**.
1. If the left child and its left grandchild are red, **rotate right**.
1. If both children are red, **flip colors**.

Each rotation or color flip is a constant-time operation, so the overall order of growth for the runtime of `add` is in *O*(log *N*).
