---
title: Unrolling recurrences
nav_order: 2
---

While this course is not focused too heavily on the mathematics behind solving recurrences, it's useful to have an intuitive understanding of a recurrence relation and be able to compute basic closed-form solutions. Recognize and apply these two identities to simplify a series.

[1 + 2 + 3 + 4 + ...](https://en.wikipedia.org/wiki/1_%2B_2_%2B_3_%2B_4_%2B_%E2%8B%AF) is approximately the square of the last term: 1 + 2 + 3 + 4 + ⋯ + *N* = *N*(*N* + 1) / 2. For a visualization of this identity, look back at our worst-case asymptotic runtime analysis of `dup1`.

[1 + 2 + 4 + 8 + ...](https://en.wikipedia.org/wiki/1_%2B_2_%2B_4_%2B_8_%2B_%E2%8B%AF) is 1 less than double the last term: 2<sup>0</sup> + 2<sup>1</sup> + ⋯ + 2<sup>*k*</sup> = 2<sup>*k* + 1</sup> − 1.

WolframAlpha can compute the asymptotic bound for any recurrence relation, e.g. [T(N) = T(N / 2) + 1](https://www.wolframalpha.com/input/?i=T%28N%29+%3D+T%28N%2F2%29+%2B+1). Did you recognize this recurrence as binary search? As you get more familiar with recurrences, you'll start to see them more and more often.

For these questions, consider the recurrence relation *T*(*N*) = *T*(*N*/2) + *cN* and *T*(1) = *d*.

{% capture question %}
Which of the following variants of `binarySearch` could have a runtime represented by the recurrence relation?

- [ ] Recursive `binarySearch` as presented earlier.
- [ ] Recursive `binarySearch` but also printing out the value of `sorted[mid]`.
- [ ] Recursive `binarySearch` but also printing out the range of values `sorted[low]` to `sorted[high]`.
- [ ] Recursive `binarySearch` but also printing out the entire `sorted` array.
{% endcapture %}
{% capture explanation %}
Options A and B are equivalent in terms of the asymptotic bound since adding another constant-time operation to print out the `sorted[mid]` value won't affect the overall asymptotic analysis.

Option C represents a program that does *N* additional work in each recursive call, where *N* represents the size of the current subproblem.

Option D represents a program that does *N* additional work in each recursive call, where *N* represents the size of the original problem.  *T*(*N*) describes the runtime for the current subproblem of size *N*, not the original problem! The overall asymptotic runtime bound for this would be *Θ*(*N*log *N*).
{% endcapture %}
{% include question.html question=question explanation=explanation number=1 %}

{% capture question %}
Unroll the recurrence and identify the big-theta asymptotic bound.

- [ ] *Θ*(1)
- [ ] *Θ*(log *N*)
- [ ] *Θ*(*N*)
- [ ] *Θ*(*N*log *N*)
{% endcapture %}
{% capture explanation %}
1. *T*(*N*) = *T*(*N*/2) + *c**N*
1. *T*(*N*) = *T*(*N*/4) + *c*(*N*/2) + *c**N*
1. *T*(*N*) = *T*(*N*/8) + *c*(*N*/4) + *c*(*N*/2) + *c**N*
1. …
1. $T(N) = d + *c*\[*N* + *N*/2 + *N*/4 + ⋯ + 1\]
1. *T*(*N*) = *d* + *c*\[2*N* − 1\] (via identity)
1. *T*(*N*) ∈ *Θ*(*N*)
{% endcapture %}
{% include question.html question=question explanation=explanation number=2 %}
